@@ -203,6 +203,8 @@ static struct Settings {
 #ifdef __linux__
     int forward_odirect;
     size_t odirect_alignment;
+
+    bool direct_io;
 #endif
 
     uid_t uid_offset;
@@ -221,13 +223,6 @@ static int is_mirroring_enabled();
 /* Checks whether the uid is to be the mirrored owner of all files. */
 static int is_mirrored_user(uid_t uid);
 
-#ifdef HAVE_FUSE_3
-#ifndef __NR_renameat2
-/* Returns the path to a temporary file inside the same directory as src_path */
-char *tmpnam_inpath(char *src_path);
-#endif
-#endif
-
 /* Processes the virtual path to a real path. Always free() the result. */
 static char *process_path(const char *path, bool resolve_symlinks);
 
@@ -377,66 +372,6 @@ static int is_mirrored_user(uid_t uid)
     return 0;
 }
 
-#ifdef HAVE_FUSE_3
-#ifndef __NR_renameat2
-char *tmpnam_inpath(char *src_path) 
-{
-    
-    char *res, tmp[L_tmpnam], *path, *fname, *tfname, *src_path_cpy, *src_path_cpy_1, *tmp_cpy;
-    int i = 0;
-    
-    src_path_cpy = (char *) calloc((strlen(src_path)), sizeof(char)); 
-    strcpy(src_path_cpy, src_path);
-    
-    src_path_cpy_1 = (char *) calloc((strlen(src_path)), sizeof(char)); 
-    strcpy(src_path_cpy_1, src_path);
-    
-    path = dirname(src_path_cpy);
-    fname = basename(src_path_cpy_1);
-    
-    tmpnam(tmp);
-    tmp_cpy = (char *) calloc((strlen(tmp)), sizeof(char));
-    strcpy(tmp_cpy, tmp);
-    
-    tfname = basename(tmp_cpy);
-    
-    res = (char *) calloc( (strlen(path) + 2 + strlen(tfname) + 1 + strlen(fname) + 1), sizeof(char));
-    
-    strcat(res, path);
-    strcat(res, "/.");
-    strcat(res, tfname);
-    strcat(res, "~");
-    strcat(res, fname);
-    
-    while (access( res, F_OK ) == 0 && i < 10) {
-        
-        tmpnam(tmp);
-        tmp_cpy = (char *) calloc((strlen(tmp)), sizeof(char));
-        strcpy(tmp_cpy, tmp);
-    
-        tfname = basename(tmp_cpy);
-    
-        res = (char *) calloc( (strlen(path) + 2 + strlen(tfname) + 1 + strlen(fname) + 1), sizeof(char));
-    
-        strcat(res, path);
-        strcat(res, "/.");
-        strcat(res, tfname);
-        strcat(res, "~");
-        strcat(res, fname);
-
-        i++;
-    }
-    
-    if (access( res, F_OK ) == 0) {
-        return NULL;
-    }
-    
-    return res;
-    
-}
-#endif
-#endif
-
 static char *process_path(const char *path, bool resolve_symlinks)
 {
     if (path == NULL) { /* possible? */
@@ -756,6 +691,9 @@ static void *bindfs_init()
     cfg->entry_timeout = 0;
     cfg->attr_timeout = 0;
     cfg->negative_timeout = 0;
+#ifdef __linux__
+    cfg->direct_io = settings.direct_io;
+#endif
     #endif
 
     assert(settings.permchain != NULL);
@@ -1349,6 +1287,9 @@ static int bindfs_open(const char *path, struct fuse_file_info *fi)
     if (!settings.forward_odirect) {
         flags &= ~O_DIRECT;
     }
+#ifndef HAVE_FUSE_3  // With FUSE 3, we set this in bindfs_init
+    fi->direct_io = settings.direct_io;
+#endif
 #endif
 
     fd = open(real_path, flags);
@@ -1754,8 +1695,10 @@ static void print_usage(const char *progname)
            "  -M      --mirror-only=... Like --mirror but disallow access for\n"
            "                            all other users.\n"
            " --map=user1/user2:...      Let user2 see files of user1 as his own.\n"
-           " --map-passwd=<passwdfile>  Load uid mapping from <passwdfile>.\n"
-           " --map-group=<groupfile>    Load gid mapping from <groupfile>.\n"
+           " --map-passwd=...           Load uid mapping from passwd-like file.\n"
+           " --map-group=...            Load gid mapping from group-like file.\n"
+           " --map-passwd-rev=...       Load reversed uid mapping from  passwd-like file.\n"
+           " --map-group-rev=...        Load reversed gid mapping from group-like file.\n"
            " --uid-offset=...           Set file uid = uid + offset.\n"
            " --gid-offset=...           Set file gid = gid + offset.\n"
            "\n"
@@ -1813,7 +1756,7 @@ static void print_usage(const char *progname)
            "  --block-devices-as-files  Show block devices as regular files.\n"
            "  --multithreaded           Enable multithreaded mode. See man page\n"
            "                            for security issue with current implementation.\n"
-           "  --forward-odirect=...      Forward O_DIRECT (it's cleared by default).\n"
+           "  --forward-odirect=...     Forward O_DIRECT (it's cleared by default).\n"
            "\n"
            "FUSE options:\n"
            "  -o opt[,opt,...]          Mount options.\n"
@@ -1857,7 +1800,8 @@ enum OptionKey {
     OPTKEY_ENABLE_IOCTL,
     OPTKEY_HIDE_HARD_LINKS,
     OPTKEY_RESOLVE_SYMLINKS,
-    OPTKEY_BLOCK_DEVICES_AS_FILES
+    OPTKEY_BLOCK_DEVICES_AS_FILES,
+    OPTKEY_NO_DIRECT_IO
 };
 
 static int process_option(void *data, const char *arg, int key,
@@ -1965,7 +1909,11 @@ static int process_option(void *data, const char *arg, int key,
     case OPTKEY_BLOCK_DEVICES_AS_FILES:
         settings.block_devices_as_files = 1;
         return 0;
-
+#ifdef __linux__
+    case OPTKEY_NO_DIRECT_IO:
+        settings.direct_io = false;
+        return 0;
+#endif
     case OPTKEY_NONOPTION:
         if (!settings.mntsrc) {
             if (strncmp(arg, "/proc/", strlen("/proc/")) == 0) {
@@ -2380,6 +2328,8 @@ int main(int argc, char *argv[])
         char *map;
         char *map_passwd;
         char *map_group;
+        char *map_passwd_rev;
+        char *map_group_rev;
         char *read_rate;
         char *write_rate;
         char *create_for_user;
@@ -2424,6 +2374,8 @@ int main(int argc, char *argv[])
         OPT_OFFSET2("--map=%s", "map=%s", map, -1),
         OPT_OFFSET2("--map-passwd=%s", "map-passwd=%s", map_passwd, -1),
         OPT_OFFSET2("--map-group=%s", "map-group=%s", map_group, -1),
+        OPT_OFFSET2("--map-passwd-rev=%s", "map-passwd-rev=%s", map_passwd_rev, -1),
+        OPT_OFFSET2("--map-group-rev=%s", "map-group-rev=%s", map_group_rev, -1),
         OPT_OFFSET3("-n", "--no-allow-other", "no-allow-other", no_allow_other, -1),
 
         OPT_OFFSET2("--read-rate=%s", "read-rate=%s", read_rate, -1),
@@ -2455,6 +2407,9 @@ int main(int argc, char *argv[])
 
         OPT2("--delete-deny", "delete-deny", OPTKEY_DELETE_DENY),
         OPT2("--rename-deny", "rename-deny", OPTKEY_RENAME_DENY),
+#ifdef __linux__
+        OPT2("--no-direct-io", "no-direct-io", OPTKEY_NO_DIRECT_IO),
+#endif
 
         OPT2("--hide-hard-links", "hide-hard-links", OPTKEY_HIDE_HARD_LINKS),
         OPT2("--resolve-symlinks", "resolve-symlinks", OPTKEY_RESOLVE_SYMLINKS),
@@ -2522,6 +2477,7 @@ int main(int argc, char *argv[])
 #ifdef __linux__
     settings.forward_odirect = 0;
     settings.odirect_alignment = 0;
+    settings.direct_io = true;
 #endif
 
     atexit(&atexit_func);
@@ -2600,6 +2556,17 @@ int main(int argc, char *argv[])
         }
     }
 
+    if (od.map_passwd_rev) {
+        if (getuid() != 0) {
+            fprintf(stderr, "Error: You need to be root to use --map-passwd-rev !\n");
+            return 1;
+        }
+        if (!parse_map_file(settings.usermap_reverse, settings.usermap, od.map_passwd_rev, 0)) {
+            /* parse_map_file printed an error */
+            return 1;
+        }
+    }
+
     /* Parse group */
     if (od.map_group) {
         if (getuid() != 0) {
@@ -2612,6 +2579,16 @@ int main(int argc, char *argv[])
         }
     }
 
+    if (od.map_group_rev) {
+        if (getuid() != 0) {
+            fprintf(stderr, "Error: You need to be root to use --map-group-rev !\n");
+            return 1;
+        }
+        if (!parse_map_file(settings.usermap_reverse, settings.usermap, od.map_group_rev, 1)) {
+            /* parse_map_file printed an error */
+            return 1;
+        }
+    }
     /* Parse usermap (may overwrite values from --map-passwd and --map-group) */
     if (od.map) {
         if (getuid() != 0) {
